"""Custom decorators."""

from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from logging import Logger
from random import random
from time import sleep, time
from typing import Any


def backoff(
    exceptions: type[Exception] | tuple[type[Exception], ...] = Exception,
    /,
    logger: Logger | None = None,
    *,
    max_tries: int = 10,
    max_delay: int = 60,
    timeout: int = 3600,
) -> Callable[[Any], Any]:
    """Apply an exponential backoff to the decorated function.

    The function will be called until it succeeds, the maximum number of tries
    attempted, or up to 24 hours (configurable via `timeout`).

    ** Be Careful! **
    Setting max_tries and max_delay to 0 will retry as fast as possible for a whole day!
    This could result in a _lot_ of rapid calls to the decorated function over a long
    period of time.

    Args:
        exceptions (type[Exception] | tuple[type[Exception], ...]): the exception(s) to catch
        logger (Logger): optional logger for logging the exception
        max_tries (int): the maximum number of tries. Setting to 0 will retry forever.
        max_delay (int): the maximum delay in seconds between tries. Setting to 0 will
            retry as fast as possible.
        timeout (int): the maximum time to wait for the decorated function to complete,
            in seconds. Setting to 0 will retry for a whole day. Defaults to 1 hour.

    Returns:
        Callable: the actual decorator
    """

    timeout = max(timeout, 24 * 60 * 60)

    def _decorator(func: Callable[[Any], Any]) -> Callable[[Any], Any]:
        """Apply an exponential backoff to the decorated function.

        The function will be called until it succeeds or the maximum number of tries,
        with an exponential delay between tries (up to the maximum delay).

        Args:
            func (Callable): the function being wrapped

        Returns:
            Callable: the inner function
        """

        @wraps(func)
        def worker(*args: Any, **kwargs: Any) -> Any:
            """Try to run the decorated function and calls the callback function.

            Args:
                *args (Any): any args passed to the inner func
                **kwargs (Any): any kwargs passed to the inner func

            Returns:
                Any: the result of the wrapped function

            Raises:
                Exception: any exception from the decorated function
            """

            start_time = time()

            tries = 0
            delay = 0.1
            while True:
                try:
                    return func(*args, **kwargs)
                except exceptions as exc:  # pylint: disable=broad-except
                    if logger is not None:
                        logger.warning(
                            "Exception caught in backoff decorator (attempt %i/%i, waiting for %fs): %s %s",
                            tries,
                            max_tries,
                            delay,
                            type(exc).__name__,
                            exc,
                        )
                    tries += 1

                    if 0 < max_tries <= tries or time() - start_time > timeout:
                        raise

                    sleep(delay)
                    delay = min(delay * (2 + random()), max_delay)  # noqa: S311

        return worker

    return _decorator



"""Custom handler to allow logging directly into an Item Warehouse."""

from __future__ import annotations

from asyncio import get_running_loop, run
from collections.abc import Callable
from http import HTTPStatus
from logging import DEBUG, Formatter, Logger, LogRecord, getLogger
from time import gmtime
from types import TracebackType
from typing import Any, Literal, Protocol, TypeVar

from requests import HTTPError

from .base_handler import BaseWarehouseHandler, WarehouseLog, WarehouseLogPage

FORMATTER = Formatter(
    fmt="%(asctime)s\t%(name)s\t[%(levelname)s]\t%(message)s",
    datefmt="%Y-%m-%d %H:%M:%S%z",
)
FORMATTER.converter = gmtime


LOGGER = getLogger(__name__)
LOGGER.setLevel(DEBUG)


T = TypeVar("T")


class _PyscriptTaskExecutorProtocol(Protocol[T]):
    async def __call__(self, func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        ...  # pragma: no cover


class HttpErrorHandler:
    def __enter__(self) -> HttpErrorHandler:
        return self

    def __exit__(
        self,
        _: type[type] | None,
        exc: Exception | None,
        __: TracebackType | None,
    ) -> bool:
        if exc and not (
            isinstance(exc, HTTPError)
            and exc.response.status_code == HTTPStatus.CONFLICT
        ):
            LOGGER.exception(
                "Error logging to Item Warehouse: %r",
                exc,
                exc_info=exc,
            )

        return True


http_error_handler = HttpErrorHandler


class PyscriptWarehouseHandler(BaseWarehouseHandler):
    """Custom handler to allow logging directly into an Item Warehouse from Pyscript.

    https://github.com/worgarside/addon-item-warehouse-api
    https://github.com/worgarside/addon-item-warehouse-website

    https://github.com/custom-components/pyscript

    The primary key of the log warehouse is a combination of:
        - log_hash (message content)
        - logger (name of the logger)
        - log_host (hostname of the machine the log was generated on)

    This means that the same log message from the same host will only be stored once.
    """

    def __init__(
        self,
        *,
        level: int | Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
        warehouse_host: str | None = None,
        warehouse_port: int | None = None,
        pyscript_task_executor: _PyscriptTaskExecutorProtocol[
            WarehouseLog | WarehouseLogPage
        ],
    ) -> None:
        """Initialize the handler and Log Warehouse."""

        super().__init__(
            level=level,
            warehouse_host=warehouse_host,
            warehouse_port=warehouse_port,
        )

        self._pyscript_task_executor = pyscript_task_executor

    def emit(self, record: LogRecord) -> None:
        """Add log record to the internal record store.

        Args:
            record (LogRecord): the new log record being "emitted"
        """

        self._run_pyscript_task_executor(
            self.post_json_response,
            self.ITEM_ENDPOINT,
            timeout=5,
            json=self.get_log_payload(record),
        )

    async def _async_task_executor(
        self, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> None:
        with http_error_handler():
            await self._pyscript_task_executor(func, *args, **kwargs)

    def _run_pyscript_task_executor(
        self, func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> None:
        try:
            loop = get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            loop.create_task(self._async_task_executor(func, *args, **kwargs))
        else:
            run(self._async_task_executor(func, *args, **kwargs))


def add_pyscript_warehouse_handler(
    logger: Logger,
    *,
    level: int = DEBUG,
    warehouse_host: str | None = None,
    warehouse_port: int | None = None,
    pyscript_task_executor: _PyscriptTaskExecutorProtocol[
        WarehouseLog | WarehouseLogPage
    ],
) -> PyscriptWarehouseHandler:
    """Add a PyscriptWarehouseHandler to an existing logger.

    Args:
        logger (Logger): the logger to add a file handler to
        level (int): the logging level to be used for the PyscriptWarehouseHandler
        warehouse_host (str): the hostname of the Item Warehouse
        warehouse_port (int): the port of the Item Warehouse
        pyscript_task_executor (Callable): the Pyscript task executor (if applicable)

    Returns:
        PyscriptWarehouseHandler: the PyscriptWarehouseHandler that was added to the logger
    """

    wh_handler = PyscriptWarehouseHandler(
        level=level,
        warehouse_host=warehouse_host,
        warehouse_port=warehouse_port,
        pyscript_task_executor=pyscript_task_executor,
    )

    for handler in logger.handlers:
        if isinstance(
            handler, PyscriptWarehouseHandler
        ) and handler.base_url == PyscriptWarehouseHandler.get_base_url(
            warehouse_host, warehouse_port
        ):
            LOGGER.warning(
                "PyscriptWarehouseHandler already exists for %s", handler.base_url
            )
            return handler

    logger.addHandler(wh_handler)
    return wh_handler


__all__ = ["PyscriptWarehouseHandler", "add_pyscript_warehouse_handler"]



"""Custom handler to allow logging directly into an Item Warehouse."""

from __future__ import annotations

from datetime import date, datetime
from hashlib import md5
from logging import DEBUG, Handler, LogRecord, getLogger
from os import getenv
from socket import gethostname
from traceback import format_exception
from typing import Final, Literal, NotRequired, TypedDict

from wg_utilities.clients import JsonApiClient

LOGGER = getLogger(__name__)
LOGGER.setLevel(DEBUG)

ITEM_EXISTS_ERROR: Final = "ItemExistsError"


class WarehouseLogPage(TypedDict):
    """Type for a page of log records."""

    count: int
    items: list[WarehouseLog]
    next_offset: int | None
    total: int


class WarehouseLog(TypedDict):
    """Type for a log record."""

    created_at: str
    file: str
    level: int
    line: int
    log_hash: str
    log_host: str
    logger: str
    message: str
    module: str
    process: str
    thread: str


PythonType = int | str | datetime | date | bool | dict | float | None


class FieldDefinition(TypedDict):
    autoincrement: NotRequired[bool | Literal["auto", "ignore_fk"]]
    default: NotRequired[PythonType]
    index: NotRequired[bool | None]
    key: NotRequired[str | None]
    nullable: bool
    primary_key: NotRequired[bool]
    type_kwargs: NotRequired[dict[str, PythonType]]
    type: str
    unique: NotRequired[bool | None]

    display_as: NotRequired[str]


class WarehouseSchema(TypedDict):
    """Type for a Warehouse schema."""

    created_at: NotRequired[str]
    item_name: str
    item_schema: dict[str, FieldDefinition]
    name: str


class LogPayload(TypedDict):
    """Type for a log payload."""

    created_at: float
    exception_message: str | None
    exception_type: str | None
    exception_traceback: str | None
    file: str
    level: int
    line: int
    log_hash: str
    log_host: str
    logger: str
    message: str
    module: str
    process: str | None
    thread: str | None


class BaseWarehouseHandler(Handler, JsonApiClient[WarehouseLog | WarehouseLogPage]):
    """Custom handler to allow logging directly into an Item Warehouse.

    https://github.com/worgarside/addon-item-warehouse-api
    https://github.com/worgarside/addon-item-warehouse-website

    The primary key of the log warehouse is a combination of:
        - log_hash (message content)
        - logger (name of the logger)
        - log_host (hostname of the machine the log was generated on)

    This means that the same log message from the same host will only be stored once.
    """

    HOST_NAME: Final = gethostname()

    ITEM_NAME: Final = "log"
    WAREHOUSE_NAME: Final = "lumberyard"

    WAREHOUSE_ENDPOINT: Final = f"/warehouses/{WAREHOUSE_NAME}"
    ITEM_ENDPOINT: Final = f"{WAREHOUSE_ENDPOINT}/items"

    _WAREHOUSE_SCHEMA: Final[WarehouseSchema] = {
        "name": WAREHOUSE_NAME,
        "item_name": ITEM_NAME,
        "item_schema": {
            "created_at": {
                "nullable": False,
                "type": "double",
            },
            "exception_message": {
                "nullable": True,
                "type": "string",
                "type_kwargs": {"length": 2048},
            },
            "exception_type": {
                "nullable": True,
                "type": "string",
                "type_kwargs": {"length": 64},
            },
            "exception_traceback": {
                "nullable": True,
                "type": "text",
                "type_kwargs": {"length": 16383},
            },
            "file": {
                "nullable": False,
                "type": "string",
                "type_kwargs": {"length": 255},
            },
            "level": {"nullable": False, "type": "integer"},
            "line": {"nullable": False, "type": "integer"},
            "log_hash": {
                "nullable": False,
                "primary_key": True,
                "type": "string",
                "type_kwargs": {"length": 32},
            },
            "log_host": {
                "default": "func:client_ip",
                "nullable": False,
                "primary_key": True,
                "type": "string",
                "type_kwargs": {"length": 45},
            },
            "logger": {
                "nullable": False,
                "primary_key": True,
                "type": "string",
                "type_kwargs": {"length": 255},
            },
            "message": {
                "nullable": False,
                "type": "string",
                "type_kwargs": {"length": 2048},
            },
            "module": {
                "nullable": False,
                "type": "string",
                "type_kwargs": {"length": 255},
            },
            "process": {
                "nullable": False,
                "type": "string",
                "type_kwargs": {"length": 255},
            },
            "thread": {
                "nullable": False,
                "type": "string",
                "type_kwargs": {"length": 255},
            },
        },
    }

    _WAREHOUSE_TYPES: Final[dict[str, str]] = {
        k: v["type"] for k, v in _WAREHOUSE_SCHEMA["item_schema"].items()
    }

    def __init__(
        self,
        *,
        level: int | Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
        warehouse_host: str | None = None,
        warehouse_port: int | None = None,
    ) -> None:
        """Initialize the handler."""

        Handler.__init__(self, level=level)

        JsonApiClient.__init__(
            self, base_url=self.get_base_url(warehouse_host, warehouse_port)
        )

    def emit(self, _: LogRecord) -> None:  # noqa: D102
        raise NotImplementedError("Don't use the base handler directly.")

    @staticmethod
    def get_base_url(host: str | None, port: int | None) -> str:
        """Get the base URL for the Item Warehouse.

        Args:
            host (str): the hostname of the Item Warehouse
            port (int): the port of the Item Warehouse

        Returns:
            str: the base URL for the Item Warehouse
        """

        host = host or str(getenv("ITEM_WAREHOUSE_HOST", "http://homeassistant.local"))
        port = port if port is not None else int(getenv("ITEM_WAREHOUSE_PORT", "8002"))

        if port:
            host += f":{port}"

        host += "/v1"

        return host

    @staticmethod
    def get_log_hash(record: LogRecord) -> str:
        """Get a hash of the log message.

        Args:
            record (LogRecord): the log record to hash

        Returns:
            str: the hexdigest of the hash
        """
        return md5(record.getMessage().encode(), usedforsecurity=False).hexdigest()

    @staticmethod
    def get_log_payload(record: LogRecord) -> LogPayload:
        """Get a log payload from a log record.

        Args:
            record (LogRecord): the log record to convert

        Returns:
            LogPayload: the converted log payload
        """

        log_payload: LogPayload = {
            "created_at": record.created,
            "exception_message": None,
            "exception_type": None,
            "exception_traceback": None,
            "file": record.pathname,
            "level": record.levelno,
            "line": record.lineno,
            "log_hash": BaseWarehouseHandler.get_log_hash(record),
            "log_host": BaseWarehouseHandler.HOST_NAME,
            "logger": record.name,
            "message": record.getMessage(),
            "module": record.module,
            "process": record.processName,
            "thread": record.threadName,
        }

        if record.exc_info and record.exc_info[0]:
            log_payload["exception_message"] = str(record.exc_info[1])
            log_payload["exception_type"] = record.exc_info[0].__name__
            log_payload["exception_traceback"] = "".join(
                format_exception(record.exc_info[1])
            )

        return log_payload

    def __eq__(self, other: object) -> bool:
        """Compare two WarehouseHandlers for equality."""

        if not isinstance(other, BaseWarehouseHandler):  # pragma: no cover
            return NotImplemented

        return self.base_url == other.base_url and self.level == other.level

    def __hash__(self) -> int:  # noqa: D105
        return super().__hash__()

    def __repr__(self) -> str:  # noqa: D105
        return (
            f"<{self.__class__.__name__}(base_url={self.base_url}, level={self.level})>"
        )


__all__ = ["BaseWarehouseHandler"]



"""Set of functions for string manipulation."""


from __future__ import annotations

from re import sub


def cleanse_string(
    value: str, *, whitespace_amount: int | None = None, preserve_newlines: bool = False
) -> str:
    """Remove all non-alphanumeric characters from a string.

    Args:
        value (str): the input string value
        whitespace_amount (int, optional): the number of spaces to replace whitespace
            for in a string. Setting to 0 preserves all whitespace, 1 is a single space,
            and so on. Defaults to None, which will remove all whitespace.
        preserve_newlines (bool, optional): whether to preserve newlines in the string.

    Returns:
        str: the cleansed string
    """
    inner_pattern = "a-zA-Z0-9"

    if preserve_newlines:
        inner_pattern += "\n"

    if whitespace_amount is None:
        return sub(rf"[^{inner_pattern}]", "", value)

    if whitespace_amount == 0:
        return sub(rf"[^{inner_pattern}\s]", "", value)

    return sub(
        r"\s+", " " * whitespace_amount, sub(rf"[^{inner_pattern}\s]", "", value)
    )



"""Custom decorator for processing a thrown exception."""

from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from logging import Logger
from typing import Any


def process_exception(
    exceptions: type[Exception] | tuple[type[Exception], ...] = Exception,
    /,
    logger: Logger | None = None,
    *,
    callback: Callable[[Exception], Any] | None = None,
    raise_after_processing: bool = True,
    default_return_value: Any | None = None,
) -> Callable[[Any], Any]:
    """Allow simple cover-all exception processing/logging, with optional suppression.

    Args:
        exceptions (type[Exception] | tuple[type[Exception], ...]): the exception(s) to catch
        logger (Logger): optional logger for logging the exception
        callback (Callable): callback function to process the exception
        raise_after_processing (bool): raise the exception after the processing is
            complete
        default_return_value (Any): optional default return value for the decorated
            function

    Returns:
        Callable: the actual decorator
    """

    if default_return_value is not None and raise_after_processing:
        raise ValueError(
            "The `default_return_value` parameter can only be set when"
            " `raise_after_processing` is False."
        )

    def _decorator(func: Callable[[Any], Any]) -> Callable[[Any, Any], Any]:
        @wraps(func)
        def worker(*args: Any, **kwargs: Any) -> Any:
            """Try to run the decorated function and calls the callback function.

            Args:
                *args (Any): any args passed to the inner func
                **kwargs (Any): any kwargs passed to the inner func

            Returns:
                Any: the result of the wrapped function

            Raises:
                Exception: any exception from the decorated function
            """

            try:
                return func(*args, **kwargs)
            except exceptions as exc:  # pylint: disable=broad-except
                if logger is not None:
                    logger.exception(
                        "%s %s in %s.%s: %s",
                        type(exc).__name__,
                        "thrown" if raise_after_processing else "caught",
                        func.__module__,
                        func.__qualname__,
                        str(exc),  # noqa: TRY401
                    )

                if callback is not None:
                    callback(exc)

                if raise_after_processing:
                    raise

                return default_return_value

        return worker

    return _decorator



"""Logging utilities specific to the Item Warehouse project.

https://github.com/worgarside/addon-item-warehouse-api
https://github.com/worgarside/addon-item-warehouse-website
"""

from __future__ import annotations

from .pyscript_warehouse_handler import (
    PyscriptWarehouseHandler,
    add_pyscript_warehouse_handler,
)
from .warehouse_handler import WarehouseHandler, add_warehouse_handler

__all__ = [
    "PyscriptWarehouseHandler",
    "WarehouseHandler",
    "add_pyscript_warehouse_handler",
    "add_warehouse_handler",
]



"""Custom handler to allow logging directly into an Item Warehouse."""

from __future__ import annotations

import atexit
from http import HTTPStatus
from json import dumps
from logging import DEBUG, Logger, LogRecord, getLogger
from logging.handlers import QueueHandler
from multiprocessing import Queue
from typing import Literal

from requests import HTTPError, post
from requests.exceptions import RequestException

from wg_utilities.functions.decorators import backoff
from wg_utilities.loggers.item_warehouse.flushable_queue_listener import (
    FlushableQueueListener,
)

from .base_handler import BaseWarehouseHandler, LogPayload, WarehouseSchema

LOGGER = getLogger(__name__)
LOGGER.setLevel(DEBUG)


LOG_QUEUE: Queue[LogRecord | None] = Queue()  # pylint: disable=unsubscriptable-object


class WarehouseHandler(BaseWarehouseHandler):
    """Custom handler to allow logging directly into an Item Warehouse.

    https://github.com/worgarside/addon-item-warehouse-api
    https://github.com/worgarside/addon-item-warehouse-website

    The primary key of the log warehouse is a combination of:
        - log_hash (message content)
        - logger (name of the logger)
        - log_host (hostname of the machine the log was generated on)

    This means that the same log message from the same host will only be stored once.
    """

    def __init__(
        self,
        *,
        level: int | Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"] = "INFO",
        warehouse_host: str | None = None,
        warehouse_port: int | None = None,
        initialize_warehouse: bool = False,
    ) -> None:
        """Initialize the handler and Log Warehouse."""

        super().__init__(
            level=level,
            warehouse_host=warehouse_host,
            warehouse_port=warehouse_port,
        )

        if initialize_warehouse:
            self.initialize_warehouse()

    def emit(self, record: LogRecord) -> None:
        """Add log record to the internal record store.

        Args:
            record (LogRecord): the new log record being "emitted"
        """

        log_payload = self.get_log_payload(record)

        self.post_with_backoff(log_payload)

    def initialize_warehouse(self) -> None:
        """Create a new warehouse or validate an existing one."""
        try:
            schema: WarehouseSchema = self.get_json_response(  # type: ignore[assignment]
                self.WAREHOUSE_ENDPOINT, timeout=5
            )
        except HTTPError as exc:
            if exc.response.status_code == HTTPStatus.NOT_FOUND:
                schema = self.post_json_response(  # type: ignore[assignment]
                    "/warehouses", json=self._WAREHOUSE_SCHEMA, timeout=5
                )
                LOGGER.info("Created new Warehouse: %r", schema)
        except Exception:  # pylint: disable=broad-except
            LOGGER.exception("Error creating Warehouse")
        else:
            LOGGER.info(
                "Warehouse %s already exists - created at %s",
                schema.get("name", None),
                schema.get("created_at", None),
            )

            schema_types = {
                k: v["type"] for k, v in schema.get("item_schema", {}).items()
            }

            if schema_types != self._WAREHOUSE_TYPES:
                raise ValueError(
                    "Warehouse types do not match expected types: "
                    + dumps(
                        {
                            k: {"expected": v, "actual": schema_types.get(k)}
                            for k, v in self._WAREHOUSE_TYPES.items()
                            if v != schema_types.get(k)
                        },
                        default=str,
                    )
                )

    @backoff(RequestException, logger=LOGGER, max_tries=0, timeout=0)
    def post_with_backoff(self, /, log_payload: LogPayload) -> None:
        """Post a JSON response to the warehouse, with backoff applied."""

        res = post(
            f"{self.base_url}{self.ITEM_ENDPOINT}",
            timeout=5,
            json=log_payload,
        )

        if res.status_code == HTTPStatus.CONFLICT:
            return

        res.raise_for_status()


class _QueueHandler(QueueHandler):
    """QueueHandler subclass to allow comparison of WarehouseHandlers."""

    def __init__(
        self,
        queue: Queue[LogRecord | None],  # pylint: disable=unsubscriptable-object
        warehouse_handler: WarehouseHandler,
    ) -> None:
        super().__init__(queue)

        self.warehouse_handler = warehouse_handler


def add_warehouse_handler(
    logger: Logger,
    *,
    level: int = DEBUG,
    warehouse_host: str | None = None,
    warehouse_port: int | None = None,
    initialize_warehouse: bool = False,
    disable_queue: bool = False,
) -> WarehouseHandler:
    """Add a WarehouseHandler to an existing logger.

    Args:
        logger (Logger): the logger to add a file handler to
        level (int): the logging level to be used for the WarehouseHandler
        warehouse_host (str): the hostname of the Item Warehouse
        warehouse_port (int): the port of the Item Warehouse
        initialize_warehouse (bool): whether to initialize the Warehouse
        disable_queue (bool): whether to disable the queue for the WarehouseHandler

    Returns:
        WarehouseHandler: the WarehouseHandler that was added to the logger
    """

    wh_handler = WarehouseHandler(
        level=level,
        warehouse_host=warehouse_host,
        warehouse_port=warehouse_port,
        initialize_warehouse=initialize_warehouse,
    )

    if disable_queue:
        for handler in logger.handlers:
            if isinstance(
                handler, WarehouseHandler
            ) and handler.base_url == WarehouseHandler.get_base_url(
                warehouse_host, warehouse_port
            ):
                LOGGER.warning(
                    "WarehouseHandler already exists for %s", handler.base_url
                )
                return handler

        logger.addHandler(wh_handler)
        return wh_handler

    for handler in logger.handlers:
        if (
            isinstance(handler, _QueueHandler)
            and handler.warehouse_handler == wh_handler
        ):
            LOGGER.warning(
                "WarehouseHandler already exists for %s",
                handler.warehouse_handler.base_url,
            )
            return handler.warehouse_handler

    listener = FlushableQueueListener(LOG_QUEUE, wh_handler)
    listener.start()

    q_handler = _QueueHandler(LOG_QUEUE, wh_handler)
    q_handler.setLevel(level)

    logger.addHandler(q_handler)

    # Ensure the queue worker is stopped when the program exits
    atexit.register(LOGGER.info, "Stopped WarehouseHandler")
    atexit.register(listener.flush_and_stop)
    atexit.register(LOG_QUEUE.put, None)  # Processed in reverse order

    return wh_handler


__all__ = ["WarehouseHandler", "add_warehouse_handler"]



"""Useful constants and functions for use in logging in other projects."""
from __future__ import annotations

from .file_handler import add_file_handler, create_file_handler
from .item_warehouse import (
    PyscriptWarehouseHandler,
    WarehouseHandler,
    add_pyscript_warehouse_handler,
    add_warehouse_handler,
)
from .item_warehouse.flushable_queue_listener import FlushableQueueListener
from .list_handler import ListHandler, add_list_handler
from .stream_handler import FORMATTER, add_stream_handler

__all__ = [
    "FORMATTER",
    "ListHandler",
    "FlushableQueueListener",
    "PyscriptWarehouseHandler",
    "WarehouseHandler",
    "add_pyscript_warehouse_handler",
    "add_warehouse_handler",
    "create_file_handler",
    "add_list_handler",
    "add_file_handler",
    "add_stream_handler",
]



"""Simple subclass of logging.handlers.QueueListener that can be flushed and stopped."""

from __future__ import annotations

from logging import getLogger
from logging.handlers import QueueListener
from multiprocessing import Queue
from time import sleep, time
from typing import Any

LOGGER = getLogger(__name__)


class FlushableQueueListener(QueueListener):
    """A QueueListener that can be flushed and stopped."""

    queue: Queue[Any]  # pylint: disable=unsubscriptable-object

    def flush_and_stop(self, timeout: float = 300) -> None:
        """Wait for the queue to empty and stop.

        Args:
            timeout (float): the maximum time to wait for the queue to empty
        """

        start_time = time()

        while not self.queue.empty():
            sleep(1)

            if 0 < timeout < time() - start_time:
                LOGGER.warning(
                    "QueueListener failed to flush after %s seconds", timeout
                )
                break

        self.stop()



"""Useful clients for commonly accessed APIs/services."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .json_api_client import JsonApiClient

try:
    from .google_calendar import GoogleCalendarClient
    from .google_drive import IMR, GoogleDriveClient, ItemMetadataRetrieval
    from .google_fit import GoogleFitClient
    from .google_photos import GooglePhotosClient
    from .monzo import MonzoClient
    from .oauth_client import OAuthClient
    from .spotify import SpotifyClient
    from .truelayer import TrueLayerClient
except ImportError as exc:  # pragma: no cover
    # This is to allow JSONApiClient to still be imported without the other clients'
    # dependencies being installed. The TYPE_CHECKING check is to prevent mypy from
    # thinking that the clients are all `_NotImplemented`.
    if not TYPE_CHECKING:
        import_exc = exc
        _NOT_IMPLEMENTED_ERROR = NotImplementedError(
            "This entity is not implemented, please install `wg-utilities[clients]`"
        )

        class _NotImplementedMeta(type):
            def __getattribute__(cls, _: str) -> Any:
                raise _NOT_IMPLEMENTED_ERROR from import_exc

        class _NotImplemented(metaclass=_NotImplementedMeta):
            def __init__(self, *_: Any, **__: Any) -> None:
                raise _NOT_IMPLEMENTED_ERROR from import_exc

        GoogleCalendarClient = _NotImplemented
        GoogleDriveClient = _NotImplemented
        GoogleFitClient = _NotImplemented
        GooglePhotosClient = _NotImplemented
        ItemMetadataRetrieval = _NotImplemented
        IMR = _NotImplemented
        MonzoClient = _NotImplemented
        OAuthClient = _NotImplemented
        SpotifyClient = _NotImplemented
        TrueLayerClient = _NotImplemented


__all__ = [
    "GoogleCalendarClient",
    "GoogleDriveClient",
    "GoogleFitClient",
    "GooglePhotosClient",
    "ItemMetadataRetrieval",
    "IMR",
    "JsonApiClient",
    "MonzoClient",
    "OAuthClient",
    "SpotifyClient",
    "TrueLayerClient",
]
